@@ -36,11 +36,13 @@ static const MunitTest tests[] = {
   munit_void_test(test_ngtcp2_ksl_update_key_range),
   munit_void_test(test_ngtcp2_ksl_dup),
   munit_void_test(test_ngtcp2_ksl_remove_hint),
+  munit_void_test(test_ngtcp2_ksl_remove),
   munit_test_end(),
 };
 
 const MunitSuite ksl_suite = {
-  "/ksl", tests, NULL, 1, MUNIT_SUITE_OPTION_NONE,
+  .prefix = "/ksl",
+  .tests = tests,
 };
 
 void test_ngtcp2_ksl_insert(void) {
@@ -123,7 +125,7 @@ void test_ngtcp2_ksl_insert(void) {
 
   ngtcp2_ksl_free(&ksl);
 
-  /* check merge node (head) */
+  /* check merge node (root) */
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_uint64_less, ngtcp2_ksl_uint64_less_search,
                   sizeof(uint64_t), mem);
 
@@ -132,7 +134,7 @@ void test_ngtcp2_ksl_insert(void) {
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &k, NULL));
   }
 
-  /* Removing these 3 nodes kicks merging 2 nodes under head */
+  /* Removing these 3 nodes kicks merging 2 nodes under root */
   k = 15;
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &k));
   k = 14;
@@ -140,11 +142,11 @@ void test_ngtcp2_ksl_insert(void) {
   k = 13;
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &k));
 
-  assert_size(29, ==, ksl.head->n);
+  assert_size(29, ==, ksl.root->n);
 
   ngtcp2_ksl_free(&ksl);
 
-  /* check merge node (non head) */
+  /* check merge node (non root) */
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_uint64_less, ngtcp2_ksl_uint64_less_search,
                   sizeof(uint64_t), mem);
 
@@ -161,9 +163,9 @@ void test_ngtcp2_ksl_insert(void) {
   k = 13;
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &k));
 
-  assert_size(2, ==, ksl.head->n);
-  assert_size(29, ==, ngtcp2_ksl_nth_node(&ksl, ksl.head, 0)->blk->n);
-  assert_size(18, ==, ngtcp2_ksl_nth_node(&ksl, ksl.head, 1)->blk->n);
+  assert_size(2, ==, ksl.root->n);
+  assert_size(29, ==, ksl.root->nodes[0].blk->n);
+  assert_size(18, ==, ksl.root->nodes[1].blk->n);
 
   ngtcp2_ksl_free(&ksl);
 
@@ -190,7 +192,7 @@ void test_ngtcp2_ksl_insert(void) {
     assert_uint64((uint64_t)i--, ==, *(uint64_t *)ngtcp2_ksl_it_key(&it));
   }
 
-  /* head only */
+  /* root only */
   for (i = 50; i < 88; ++i) {
     k = (uint64_t)i;
     assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &k));
@@ -204,6 +206,34 @@ void test_ngtcp2_ksl_insert(void) {
   }
 
   ngtcp2_ksl_free(&ksl);
+
+  /* Set key after split */
+  ngtcp2_ksl_init(&ksl, ngtcp2_ksl_uint64_less, ngtcp2_ksl_uint64_less_search,
+                  sizeof(uint64_t), mem);
+
+  /* split nodes */
+  for (i = 0; i < 46; ++i) {
+    k = 90 - (uint64_t)i * 2;
+    assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &k, NULL));
+  }
+
+  k = 60;
+
+  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &k));
+
+  k = 59;
+
+  assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &k, NULL));
+
+  k = 60;
+
+  assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &k, NULL));
+
+  it = ngtcp2_ksl_lower_bound(&ksl, &k);
+
+  assert_uint64(60, ==, *(uint64_t *)ngtcp2_ksl_it_key(&it));
+
+  ngtcp2_ksl_free(&ksl);
 }
 
 void test_ngtcp2_ksl_clear(void) {
@@ -216,6 +246,8 @@ void test_ngtcp2_ksl_clear(void) {
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_uint64_less, ngtcp2_ksl_uint64_less_search,
                   sizeof(uint64_t), mem);
 
+  ngtcp2_ksl_clear(&ksl);
+
   for (i = 0; i < 100; ++i) {
     k = (uint64_t)i;
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &k, NULL));
@@ -277,51 +309,42 @@ void test_ngtcp2_ksl_range(void) {
 
   ngtcp2_ksl_free(&ksl);
 
-  /* check merge node (head) */
+  /* check merge node (root) */
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_range_compar, ngtcp2_ksl_range_search,
                   sizeof(ngtcp2_range), mem);
 
-  for (i = 0; i < 32; ++i) {
+  for (i = 0; i < 33; ++i) {
     ngtcp2_range_init(&r, i, i + 1);
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &r, NULL));
   }
 
-  /* Removing these 3 nodes kicks merging 2 nodes under head */
+  /* Removing these 2 nodes kicks merging 2 nodes under root */
   ngtcp2_range_init(&r, 13, 14);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
 
   ngtcp2_range_init(&r, 14, 15);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
 
-  ngtcp2_range_init(&r, 15, 16);
-  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
-
-  assert_size(29, ==, ksl.head->n);
+  assert_size(31, ==, ksl.root->n);
 
   ngtcp2_ksl_free(&ksl);
 
-  /* check merge node (non head) */
+  /* check merge node (non root) */
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_range_compar, ngtcp2_ksl_range_search,
                   sizeof(ngtcp2_range), mem);
 
-  for (i = 0; i < 32 + 18; ++i) {
+  for (i = 0; i < 33 + 18; ++i) {
     ngtcp2_range_init(&r, i, i + 1);
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &r, NULL));
   }
 
-  /* Removing these 3 nodes kicks merging 2 nodes */
+  /* Removing this node kicks merging 2 nodes */
   ngtcp2_range_init(&r, 13, 14);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
 
-  ngtcp2_range_init(&r, 14, 15);
-  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
-
-  ngtcp2_range_init(&r, 15, 16);
-  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
-
-  assert_size(2, ==, ksl.head->n);
-  assert_size(29, ==, ngtcp2_ksl_nth_node(&ksl, ksl.head, 0)->blk->n);
-  assert_size(18, ==, ngtcp2_ksl_nth_node(&ksl, ksl.head, 1)->blk->n);
+  assert_size(2, ==, ksl.root->n);
+  assert_size(31, ==, ksl.root->nodes[0].blk->n);
+  assert_size(19, ==, ksl.root->nodes[1].blk->n);
 
   ngtcp2_ksl_free(&ksl);
 
@@ -334,16 +357,13 @@ void test_ngtcp2_ksl_range(void) {
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &r, NULL));
   }
 
-  ngtcp2_range_init(&r, 1501, 1502);
-  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
-  ngtcp2_range_init(&r, 1401, 1402);
+  ngtcp2_range_init(&r, 3101, 3102);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
 
-  r = *(ngtcp2_range *)(void *)ngtcp2_ksl_nth_node(
-         &ksl, ngtcp2_ksl_nth_node(&ksl, ksl.head, 1)->blk, 0)
-         ->key;
+  r =
+    *(ngtcp2_range *)(void *)ngtcp2_ksl_blk_nth_key(ksl.root->nodes[1].blk, 0);
 
-  assert_uint64(1701, ==, r.begin);
+  assert_uint64(1601, ==, r.begin);
 
   ngtcp2_ksl_free(&ksl);
 
@@ -351,22 +371,23 @@ void test_ngtcp2_ksl_range(void) {
   ngtcp2_ksl_init(&ksl, ngtcp2_ksl_range_compar, ngtcp2_ksl_range_search,
                   sizeof(ngtcp2_range), mem);
 
-  for (i = 0; i < 32; ++i) {
+  for (i = 0; i < 49; ++i) {
     ngtcp2_range_init(&r, i, i + 1);
     assert_int(0, ==, ngtcp2_ksl_insert(&ksl, NULL, &r, NULL));
   }
 
-  ngtcp2_range_init(&r, 17, 18);
+  ngtcp2_range_init(&r, 15, 16);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
-  ngtcp2_range_init(&r, 16, 17);
+  ngtcp2_range_init(&r, 48, 49);
+  assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
+  ngtcp2_range_init(&r, 47, 48);
   assert_int(0, ==, ngtcp2_ksl_remove(&ksl, NULL, &r));
 
-  node = ngtcp2_ksl_nth_node(&ksl, ksl.head, 0);
-  r = *(ngtcp2_range *)(void *)ngtcp2_ksl_nth_node(&ksl, node->blk,
-                                                   node->blk->n - 1)
-         ->key;
+  node = &ksl.root->nodes[0];
+  r = *(ngtcp2_range *)(void *)ngtcp2_ksl_blk_nth_key(node->blk,
+                                                      node->blk->n - 1);
 
-  assert_uint64(14, ==, r.begin);
+  assert_uint64(23, ==, r.begin);
 
   ngtcp2_ksl_free(&ksl);
 }
@@ -525,3 +546,15 @@ void test_ngtcp2_ksl_remove_hint(void) {
     ngtcp2_ksl_free(&ksl);
   }
 }
+
+void test_ngtcp2_ksl_remove(void) {
+  ngtcp2_ksl ksl;
+  const ngtcp2_mem *mem = ngtcp2_mem_default();
+  uint64_t key = 0;
+
+  ngtcp2_ksl_init(&ksl, ngtcp2_ksl_uint64_less, ngtcp2_ksl_uint64_less_search,
+                  sizeof(uint64_t), mem);
+
+  assert_int(NGTCP2_ERR_INVALID_ARGUMENT, ==,
+             ngtcp2_ksl_remove(&ksl, NULL, &key));
+}
